"""
Session module.

This module provides the following classes:

- Session
"""
import platform
from datetime import date
from typing import List, Optional
from urllib.parse import urlencode

import requests

from serifan import __version__, comics_list, exceptions

from .utils import list_strings_to_dates


class Session:
    """Session to request api endpoints."""

    def __init__(self) -> None:
        """Intialize a new Session."""
        self.header = {
            "User-Agent": f"Serifan/{__version__} ({platform.system()}; {platform.release()})"
        }
        self.api_url = "https://api.shortboxed.com/comics/v1/{}"

    def call(self, endpoint, params=None):
        """
        Make request for api endpoints.

        :param str endpoint: The endpoint to request information from.
        :param dict params: Parameters to add to the request.
        """
        params = {} if params is None else urlencode(params)
        url = self.api_url.format("/".join(str(e) for e in endpoint))

        try:
            response = requests.get(url, params=params, headers=self.header)
        except requests.exceptions.ConnectionError as e:
            raise exceptions.ApiError("Connection error: {}".format(repr(e)))

        if (response.status_code >= 500) and (response.status_code < 600):
            raise requests.HTTPError(f"Shortboxed Server Error: {response.status_code}")

        return response.json()

    def new_releases(self) -> comics_list.ComicsList:
        """
        Request a list of this weeks current new release comics.

        :return: A list of :class:`Comic` objects.
        :rtype: ComicsList
        """
        return comics_list.ComicsList(self.call(["new"], params={}))

    def previous_releases(self) -> comics_list.ComicsList:
        """
        Request a list of the previous weeks released comics.

        :return: A list of :class:`Comic` objects.
        :rtype: ComicsList
        """
        return comics_list.ComicsList(self.call(["previous"], params={}))

    def future_releases(self) -> comics_list.ComicsList:
        """
        Request a list of the next weeks comics.

        :return: A list of :class:`Comic` objects.
        :rtype: ComicsList
        """
        return comics_list.ComicsList(self.call(["future"], params={}))

    def release_date(self, release_date: str) -> comics_list.ComicsList:
        """
        Request comics with a specific release date.

        :param release_date: Date comics where released in iso8601 format (ie: 2016-02-17).
        :type params: str

        :return: A list of :class:`Comic` objects.
        :rtype: ComicsList
        """
        return comics_list.ComicsList(self.call(["release_date", release_date], params={}))

    def available_release_dates(self) -> List[date]:
        """Retrieve list of release dates."""
        return list_strings_to_dates(self.call(["releases", "available"]))

    def query(
        self,
        publisher: Optional[str] = None,
        title: Optional[str] = None,
        creators: Optional[str] = None,
        release_date: Optional[str] = None,
    ) -> comics_list.ComicsList:
        """
        Search for a list of comics.

        :param publisher: Publisher to search by.
        :type params: str, optional

        :param title: Title to search for.
        :type params: str, optional

        :param creators: Creator to search for.
        :type params: str, optional

        :param release_date: Date comics where released in iso8601 format (ie: 2016-02-17).
        :type params: str, optional

        :return: A list of :class:`Comic` objects.
        :rtype: ComicsList
        """
        params = {}
        if publisher:
            params["publisher"] = publisher
        if title:
            params["title"] = title
        if creators:
            params["creators"] = creators
        if release_date:
            params["release_date"] = release_date

        return comics_list.ComicsList(self.call(["query"], params=params))



"""
Test Comic module.

This module contains tests for Comic objects.
"""
import datetime
from decimal import Decimal

import pytest
import requests_mock
from requests.exceptions import HTTPError

from serifan import api, comics_list, exceptions, session, utils


def test_comics_list(comic_list_response):
    """Test for comics_list."""
    res = comics_list.ComicsList(comic_list_response)
    comic_iter = iter(res)
    assert next(comic_iter).title == "BITTER ROOT #15 CVR A GREENE (MR)"
    assert next(comic_iter).title == "BITTER ROOT #15 CVR B CONLEY CURIEL (MR)"
    assert len(res) == 2
    assert res[0].publisher == "IMAGE COMICS"
    assert res[1].title == "BITTER ROOT #15 CVR B CONLEY CURIEL (MR)"
    assert res[1].price == Decimal("3.99")
    assert res[1].release_date == datetime.date(2021, 8, 11)
    assert res[1].diamond_id == "MAY210140"


def test_available_dates(sb_dates_response, mocker):
    """Test available dates api."""
    mocker.patch.object(session.Session, "call", return_value=sb_dates_response)
    sb = api()
    result = sb.available_release_dates()
    assert len(result) == 5
    assert result[0] == datetime.date(2021, 7, 29)


def test_new_releases(comic_list_response, mocker):
    """Test new releases api."""
    mocker.patch.object(session.Session, "call", return_value=comic_list_response)
    sb = api()
    result = sb.new_releases()
    assert len(result) == 2
    assert result[0].title == "BITTER ROOT #15 CVR A GREENE (MR)"


def test_previous_releases(comic_list_response, mocker):
    """Test previous releases api."""
    mocker.patch.object(session.Session, "call", return_value=comic_list_response)
    sb = api()
    result = sb.previous_releases()
    assert len(result) == 2
    assert result[0].title == "BITTER ROOT #15 CVR A GREENE (MR)"


def test_future_releases(comic_list_response, mocker):
    """Test future releases api."""
    mocker.patch.object(session.Session, "call", return_value=comic_list_response)
    sb = api()
    result = sb.future_releases()
    assert len(result) == 2
    assert result[0].title == "BITTER ROOT #15 CVR A GREENE (MR)"


def test_query(comic_list_response, mocker):
    """Test query api."""
    mocker.patch.object(session.Session, "call", return_value=comic_list_response)
    sb = api()
    result = sb.query("Image", "WildC.A.T.S.", None, None)
    assert len(result) == 2
    assert result[0].title == "BITTER ROOT #15 CVR A GREENE (MR)"


def test_bad_response():
    """Test for a bad response."""
    with pytest.raises(exceptions.ApiError):
        comics_list.ComicsList({"comics": {"name": 1}})


def test_list_string_to_date(sb_dates_response):
    """Test the list_string_to_date function."""
    results = utils.list_strings_to_dates(sb_dates_response)
    assert len(results) == 5
    assert results[0] == datetime.date(2021, 7, 29)


def test_for_shortboxed_internal_server_error():
    """Test for Shortboxed 500 response code."""
    with requests_mock.mock() as m:
        m.get(requests_mock.ANY, status_code=500, text="Shortboxed Internal Server Error")

        with pytest.raises(HTTPError):
            sb = api()
            sb.available_release_dates()



"""
Test Init module.

This module contains tests for project init.
"""
from serifan import api, session


def test_api():
    """Test for api()."""
    sb = None
    try:
        sb = api()
    except Exception as exc:
        print("serifan.api() raised {} unexpectedly!".format(exc))

    assert sb.__class__.__name__ == session.Session.__name__



"""
ComicsList module.

This module provides the following classes:

- ComicsList
"""
from marshmallow import ValidationError

from serifan import comic, exceptions


class ComicsList:
    """The ComicsList object contains a list of `Comic` objects."""

    def __init__(self, response):
        """Initialize a new ComicsList."""
        self.comics = []

        schema = comic.ComicSchema()
        for issue_dict in response["comics"]:
            try:
                result = schema.load(issue_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error)

            self.comics.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.comics)

    def __len__(self):
        """Return the length of the object."""
        return len(self.comics)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.comics[index]



"""
Comic module.

This module provides the following classes:

- Comic
- ComicSchema
"""
from marshmallow import INCLUDE, Schema, fields, post_load, pre_load


class Comic:
    """
    The Comic object contains information for an issue.

    :param `**kwargs`: The keyword arguments is used for setting comic data.
    """

    def __init__(self, **kwargs):
        """Intialize a new Comic."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class ComicSchema(Schema):
    """
    Schema for Comic api.

    .. versionchanged:: 1.0.0
        Changed ``price`` field to a ``Decimal`` type.
    """

    publisher = fields.Str()
    description = fields.Str()
    title = fields.Str()
    price = fields.Decimal(places=2)
    creators = fields.Str()
    release_date = fields.Date(format="%Y-%m-%d")
    diamond_id = fields.Str()

    class Meta:
        """Any unknown fields will be included."""

        unknown = INCLUDE

    @pre_load
    def process_input(self, data, **kwargs):
        """Strip the dollar sign from the price before loading the data."""
        new_data = data

        if "price" in new_data:
            new_data["price"] = new_data["price"].strip("$")

        return new_data

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the comic object.

        :param data: Data from Shortboxed response.
        :return: :class:`Comic` object
        :rtype: Comic
        """
        return Comic(**data)



"""
Utils module.

This module provides the following functions:

- list_strings_to_dates()
"""
from datetime import date
from datetime import datetime as dt
from typing import List


def list_strings_to_dates(lst) -> List[date]:
    """
    Convert a list of date strings to a list of date objects.

    :param lst: A list of date strings
    :type params: list of str

    :return: A list of `datetime.date` objects
    :rtype: date
    """
    # One of the string dates in the api has a period at the end.
    return [dt.strptime(d.strip("."), "%Y-%m-%d").date() for d in lst["dates"]]
